# BubbleID
# Import libraries:
import torch, detectron2
from detectron2.utils.logger import setup_logger
setup_logger()

# import some common libraries
import numpy as np
import os, json, cv2, random

# import some common detectron2 utilities
from detectron2 import model_zoo, structures
from detectron2.engine import DefaultPredictor
from detectron2.config import get_cfg
from detectron2.utils.visualizer import Visualizer
from detectron2.data import MetadataCatalog, DatasetCatalog

import torch
torch.cuda.is_available()

from detectron2.engine import DefaultTrainer
import os
import glob

import matplotlib.pyplot as plt
from tqdm import tqdm

import os
import time
from tqdm import tqdm
import numpy as np
import cv2
import filterpy
import torch
import super_gradients as sg
import matplotlib.pyplot as plt
from .ocsort import ocsort
import colorsys    
import pandas as pd
import cv2
import matplotlib.pyplot as plt
import numpy as np
from io import BytesIO
import PIL.Image
from scipy.spatial import cKDTree

import numpy as np
import matplotlib.pyplot as plt
from scipy.fft import fft2, fftfreq
from scipy.ndimage import gaussian_filter
from tqdm import tqdm
import cv2
import numpy as np

# Define Helper Functions
def get_image_paths(directory):
    image_extensions = ['*.jpg']  # Add more extensions as needed

    image_paths = []
    for extension in image_extensions:
        pattern = os.path.join(directory, '**', extension)
        image_paths.extend(glob.glob(pattern, recursive=True))

    return sorted(image_paths)  # Sort the list of image paths alphabetically

def get_color(number):
    """ Converts an integer number to a color """
    # change these however you want to
    hue = number*30 % 180
    saturation = number*103 % 256
    value = number*50 % 256

    # expects normalized values
    color = colorsys.hsv_to_rgb(hue/179, saturation/255, value/255)

    return [int(c*255) for c in color]

# Define Function for computing iou
def iou_batch(bboxes1, bboxes2):
    """
    From SORT: Computes IOU between two bboxes in the form [x1,y1,x2,y2]
    """
    bboxes2 = np.expand_dims(bboxes2, 0)
    bboxes1 = np.expand_dims(bboxes1, 1)
    
    xx1 = np.maximum(bboxes1[..., 0], bboxes2[..., 0])
    yy1 = np.maximum(bboxes1[..., 1], bboxes2[..., 1])
    xx2 = np.minimum(bboxes1[..., 2], bboxes2[..., 2])
    yy2 = np.minimum(bboxes1[..., 3], bboxes2[..., 3])
    w = np.maximum(0., xx2 - xx1)
    h = np.maximum(0., yy2 - yy1)
    wh = w * h
    o = wh / ((bboxes1[..., 2] - bboxes1[..., 0]) * (bboxes1[..., 3] - bboxes1[..., 1])                                      
        + (bboxes2[..., 2] - bboxes2[..., 0]) * (bboxes2[..., 3] - bboxes2[..., 1]) - wh)                                              
    return(o) 
    

class DataAnalysis:
    def __init__(self,imagesfolder,videopath,savefolder,extension,modelweightsloc):
        self.imagesfolder=imagesfolder
        self.videopath=videopath
        self.savefolder=savefolder
        self.extension=extension
        self.modeldirectory, self.modelweights=os.path.split(modelweightsloc)

    
    def GenerateData(self, thres=0.5):
        directory_path=self.imagesfolder
        video_file=self.videopath
        
        # define save locations
        file_path=self.savefolder+f'/bb-Boiling-{self.extension}.txt'
        output_file_path=self.savefolder+f'/bb-Boiling-output-{self.extension}.txt'
        vapor_file=self.savefolder+f'vapor_{self.extension}.npy'
        vapor_base_file=self.savefolder+f'/vaporBase_bt-{self.extension}.npy'
        bubble_size_file=self.savefolder+f'/bubble_size_bt-{self.extension}.npy'
        bubind_file=self.savefolder+f'/bubind_{self.extension}.npy'
        frameind_file=self.savefolder+f'/frames_{self.extension}.npy'
        classind_file=self.savefolder+f'/class_{self.extension}.npy'
        bubclassind_file=self.savefolder+f'/bubclass_{self.extension}.npy'

        # Make save folder if it does not exist
        if not os.path.exists(self.savefolder):
            os.makedirs(self.savefolder)

        # load model
        print("Load model")
        cfg = get_cfg()
        cfg.OUTPUT_DIR = self.modeldirectory
        cfg.merge_from_file(model_zoo.get_config_file("COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml"))
        cfg.DATALOADER.NUM_WORKERS = 2
        cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url("COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml")  # Let training initialize from model zoo
        cfg.SOLVER.IMS_PER_BATCH = 2  # This is the real "batch size" commonly known to deep learning people
        cfg.SOLVER.BASE_LR = 0.00025  # pick a good LR
        cfg.SOLVER.MAX_ITER = 1000    # 1000 iterations seems good enough for this dataset
        cfg.SOLVER.STEPS = []        # do not decay learning rate
        cfg.MODEL.ROI_HEADS.BATCH_SIZE_PER_IMAGE = 256   # Default is 512, using 256 for this dataset.
        cfg.MODEL.ROI_HEADS.NUM_CLASSES = 2 
        cfg.MODEL.WEIGHTS = os.path.join(cfg.OUTPUT_DIR, self.modelweights)  # path to the model we just trained
        cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5   # set a custom testing threshold
        predictor = DefaultPredictor(cfg)

        # Get list of sorted image files
        print("Load image paths")
        image_paths = get_image_paths(directory_path)
        image_paths_sub=image_paths[0:]

        print("Run instance segmentation model and save data")
        Bounding_Box=np.empty((0,7))
        bubble_size=[]
        vapor=[]
        vapor_base=[]
        for i in tqdm(range(len(image_paths_sub))):
        #for i in [117,118,119]:    
            new_im = cv2.imread(image_paths_sub[i])
            outputs  = predictor(new_im)
            box=outputs["instances"].pred_boxes
            box=box.tensor
            box=box.cpu().tolist()
            masks=outputs["instances"].pred_masks.cpu()
            scores=outputs["instances"].scores
            #scores=scores.tensor
            scores=scores.cpu().tolist()
            class_val=outputs["instances"].pred_classes.cpu().tolist()

            converted_bounding_box=[]
            if len(scores)>0:
                for j in range(len(box)):

                    x1, y1, x2, y2 = box[j]

                    #if y2 > 502 and y2<533 and x1>320 and x1<515:
                    #if y2 > 502 and y2<680 and x1>250 and x1<580:
                    if y2 > 0:

                        converted_bounding_box.append([x1,y1,x2,y2])
                    elif y2 > 502 and y2<533 and x2>320 and x2<515:
                    #elif y2 > 502 and y2<680 and x2>250 and x2<580:
                        converted_bounding_box.append([x1,y1,x2,y2])

                box=converted_bounding_box
                box_data = [[i+1]+ box[j] +[scores[j]] +[class_val[j]] for j in range(len(box))]

                if len(box_data) !=0:
                    Bounding_Box=np.vstack([Bounding_Box, box_data])
                else:
                    print('Error')
                    outputs=predictor(new_im)
                    box=outputs["instances"].pred_boxes
                    box=box.tensor
                    box=box.cpu().tolist()
                    x1,y1,x2,y2=box[0]
                    box_data = [[i+1]+ box[j] +[scores[j]]+ [class_val[j]] for j in range(len(box))]

                    Bounding_Box=np.vstack([Bounding_Box, box_data])

                masks=outputs['instances'].pred_masks.cpu()
                scores=outputs['instances'].scores.cpu()
                index_tensor=torch.tensor([k for k in range(len(masks))])
                index_to_keep=index_tensor[scores>thres]
                masks=torch.index_select(masks, 0, index_to_keep)
                class_val=np.array(class_val)[index_to_keep]
                combined_mask=torch.any(masks, axis=0)
                vapor.append(torch.sum(combined_mask).item())
                indexs=np.where(np.array(class_val)==0)[0]
                masks_base=masks[indexs]
                combined_mask=torch.any(masks_base, axis=0)
                vapor_base.append(torch.sum(combined_mask).item())
                pixel_count=torch.sum(masks, dim=(1,2)).numpy()
                bubble_size.append(pixel_count)
            else:
                print(i)



        # Save vapor and bubble size
        np.save(vapor_file, vapor)
        np.save(vapor_base_file, vapor_base)
        np.save(bubble_size_file,bubble_size)

        # Save bounding box data
        with open(file_path, 'w') as file:
            for sublist in Bounding_Box:
                formatted_values = []
                for value in sublist:
                    if isinstance(value, int):
                        formatted_values.append(f'{int(value)}')  # Format integers as strings
                    elif isinstance(value, float):
                        formatted_values.append(f'{value:.4f}')  # Format floats with 4 decimal places
                    else:
                        formatted_values.append(str(value))  # Keep other types as they are

                # Manually format the first and last elements as integers
                formatted_values[0] = str(int(float(formatted_values[0])))
                formatted_values[-1] = str(int(float(formatted_values[-1])))
                # Join the formatted values with spaces
                line = ','.join(formatted_values)

                # Write the line to the file and add a newline character
                file.write(line + '\n')

        print("Perform ocsort tracking on saved data")

        tracker=ocsort.OCSort(det_thresh=thres, max_age=10, min_hits=20)
        img_info=(600,832)
        img_size=(600,832)
        boiling_test=78
        hf=120
        #cap=cv2.VideoCapture(f"/mnt/share/zdrive/Hari/Bubble hydro acoustic analysis/B83_210_Bubble/Bubble/video.avi")
        #cap=cv2.VideoCapture(f"/mnt/share/zdrive/Christy/Boiling-{boiling_test}/{hf}W-full.avi")
        cap=cv2.VideoCapture(video_file)


        if (cap.isOpened() == False):
            print("Error opening video file")

        frames=[]
        i=0
        counter, fps, elapsed=0,0,0
        frame_data={}
        with open(file_path, 'r') as file:
            for line in file:
                # Split the line into components
                parts = line.strip().split(',')

                # Extract frame ID and data
                frame_id = int(parts[0])
                data = list(map(float, parts[1:-1]))


                # Check if the frame ID is already in the dictionary
                if frame_id not in frame_data:
                    frame_data[frame_id] = []

                # Append the data to the corresponding frame ID
                frame_data[frame_id].append(data)

        with open(output_file_path, 'w') as file:

            while(cap.isOpened()):
                ret, frame=cap.read()

                if ret==True:
                    # load data from previously configured text file that has
                    # bounding boxes. (x1,y1,x2,y2,c)
                    xyxyc=frame_data.get(i+1, [])
                    i+=1
                    if np.array(xyxyc).size==0:
                        xyxyc=np.empty((0,5))
                    tracks=tracker.update(np.array(xyxyc), img_info, img_size)
                    for track in tracker.trackers:

                        track_id=track.id
                        hits=track.hits
                        color=get_color(track_id*15)
                        x1,y1,x2,y2=np.round(track.get_state()).astype(int).squeeze()
                        '''
                        #cv2.rectangle(frame, (x1,y1),(x2,y2), color, 2)
                        #cv2.putText(frame, f"{track_id}-{hits}",
                         #           (x1+10, y1-5),
                          #          cv2.FONT_HERSHEY_SIMPLEX,
                           #         0.5,
                            #        color,
                             #       1,
                              #      cv2.LINE_AA)
                        #frames.append(frame)
                        '''
                        file.write(f'{i},{track_id},{hits},{x1},{y1},{x2},{y2}\n')
                else:
                    break
            cap.release()
            del cap


        print("Match tracking results to bubble indexs")    
        # load data
        real_data=np.loadtxt(file_path, delimiter=",")
        real=real_data[real_data[:,-2]>=thres]
        real=np.round(real[:,0:-2]).astype(int)
        #del real_data
        pred_data=np.loadtxt(output_file_path,delimiter=",").astype(int)


        # Initialize an empty dictionary
        my_dict = {}

        # Assign keys ranging from 1 to 10 using a for loop
        for i in range(1, real[-1][0]+1):
            my_dict[i] = []

        for item in real:
            key=item[0]
            value=item[1:]
            if my_dict[key] is []:
                my_dict[key]=(list(value))
            else:
                my_dict[key].append(list(value))
        realg=[values for values in my_dict.values()]

        # Initialize an empty dictionary
        my_dict = {}

        # Assign keys ranging from 1 to 10 using a for loop
        for i in range(1, real[-1][0]+1):
            my_dict[i] = []

        for item in real:
            key=item[0]
            value=item[1]
            if my_dict[key] is []:
                my_dict[key]=((value))
            else:
                my_dict[key].append((value))
        realgG=[values for values in my_dict.values()]

        # Initialize an empty dictionary
        my_dict = {}

        # Assign keys ranging from 1 to 10 using a for loop
        for i in range(1, real[-1][0]+1):
            my_dict[i] = []

        for item in pred_data:
            key=item[0]
            value=item[1:]
            if my_dict[key] is []:
                my_dict[key]=(list(value))
            else:
                my_dict[key].append(list(value))
        predg=[values for values in my_dict.values()]


        import copy
        values=copy.deepcopy(realg)
        tracks=copy.deepcopy(realgG)

        # given previous frame and current frame remove row that doesn't have 

        for k in range(len(predg)-1):
        #for k in range(2):
            if len(predg[k]) > 0 and len(predg[k+1]) >0: 
                frame1=predg[k]
                frame2=predg[k+1]
                vector1=np.array(frame2)[:,0].tolist()
                vector2=np.array(frame1)[:,0].tolist()
                result_vector = np.full(len(vector1), -1)
                for i, val in enumerate(vector1):
                    if val in vector2:
                        result_vector[i] = vector2.index(val)
                j=0
                for i in range(len(result_vector)):
                    if result_vector[i] !=-1:
                        if frame2[i][1]!= frame1[result_vector[i]][1]:
                            #print(k,j,i)
                            tracks[k+1][j]=frame2[i][0]
                            values[k+1][j][:]=frame2[i][2:]
                            j+=1
                    if result_vector[i] == -1:

                        tracks[k+1][j]=frame2[i][0]
                        values[k+1][j][:]=frame2[i][2:]
                        j+=1

        # Tracks is of shape (frames, bb in real detection (basically removes tracks with no hits))
        tracks[0]=np.array(predg[0])[:,0].tolist()
        values[0]=np.array(predg[0])[:,2:].tolist()

        for i in range(len(values)):
            if len(values[i]) >0:
                sort=np.argmax(iou_batch(realg[i],values[i]),axis=1).tolist()
                tracks[i]=np.array(tracks[i])[sort].tolist()


        # Original data
        data = tracks

        # Find the maximum number in the original data
        #max_number = max(max(row) for row in data)
        max_number = max(max(sublist, default=float('-inf')) for sublist in data if sublist)

        # Create a list of lists to store the index positions and initial row numbers
        frames = [[] for _ in range(max_number + 1)]

        # Iterate through the original data and populate the result list
        for initial_row, row in enumerate(data):
            for index, number in enumerate(row):
                frames[number].append(initial_row)


        # Original data
        data = tracks
        # Find the maximum number in the original data
        #max_number = max(max(row) for row in data)
        max_number = max(max(sublist, default=float('-inf')) for sublist in data if sublist)


        # Create a list of lists to store the index positions
        bubInd= [[] for _ in range(max_number + 1)]

        # Iterate through the original data and populate the result list
        for row in data:
            for index, number in enumerate(row):
                bubInd[number].append(index)

        # Print the result
        #print(bubInd)

        # Save Data
        bubInd=np.array(bubInd, dtype=object)
        np.save(bubind_file, bubInd)
        frames=np.array(frames, dtype=object)
        np.save(frameind_file, frames)




        classes=real_data[:,-1]

        # Initialize an empty dictionary
        my_dict = {}

        # Assign keys ranging from 1 to 10 using a for loop
        for i in range(1, real[-1][0]+1):
            my_dict[i] = []

        for item in real_data:
            key=item[0]
            value=item[-1]
            if my_dict[key] is []:
                my_dict[key]=((value))
            else:
                my_dict[key].append((value))
        realgG=[values for values in my_dict.values()]

        np.save(classind_file,realgG)

        bub_class=copy.deepcopy(frames)
        for j in range(len(bubInd)):
            bub=j
            for i in range(len(frames[bub])):
                bub_class[j][i]=realgG[frames[bub][i]][bubInd[bub][i]]


        bub_class=np.array(bub_class, dtype=object)
        np.save(bubclassind_file, bub_class)
        print("Finish")
        
    def Plotvf(self):
        vf_path=self.savefolder+f'vapor_{self.extension}.npy'
        vidstart=0
        vf=np.load(vf_path)/(832*600)
        time=[(i/150)+vidstart for i in range(len(vf))]
        df=pd.DataFrame(data=vf)
        df['value']=df.iloc[:,0].rolling(window=300).mean()
        plt.figure(figsize=(5,10))
        fig, ax1=plt.subplots(figsize=(6,2))
        ax1.plot(time, df, color='lightgray')
        ax1.plot(time, df['value'], color='darkblue', label='Rolling Average')
        ax1.set_xlabel('Time (s)')
        ax1.set_ylabel('Vapor Fraction')
        ax1.legend()
        saveloc=self.savefolder+f'vaporfig_{self.extension}.png'
        plt.savefig(saveloc, bbox_inches='tight')
        plt.show()

        
    def Plotbc(self):
        bs_path=self.savefolder+f'/bubble_size_bt-{self.extension}.npy'
        bs=np.load(bs_path, allow_pickle=True)
        count=[]
        for i in range(len(bs)):
            count.append(len(bs[i])) 
        df=pd.DataFrame(data=count)
        
        vidstart=0
        time=[(i/150)+vidstart for i in range(len(count))]
        df['value']=df.iloc[:,0].rolling(window=300).mean()
        plt.figure(figsize=(5,10))
        fig, ax1=plt.subplots(figsize=(6,2))
        ax1.plot(time, df, color='lightgray')
        ax1.plot(time, df['value'], color='darkblue', label='Rolling Average')
        ax1.set_xlabel('Time (s)')
        ax1.set_ylabel('Bubble Count')
        ax1.legend()
        saveloc = self.savefolder + f'bcfig_{self.extension}.png'
        plt.savefig(saveloc, bbox_inches='tight')
        plt.show()
        
        
    def PlotInterfaceVelocity(self, bubble):
        directory_path=self.imagesfolder

        bubind_file=self.savefolder+f'/bubind_{self.extension}.npy'
        frameind_file=self.savefolder+f'/frames_{self.extension}.npy'


        # load model
        cfg = get_cfg()
        cfg.OUTPUT_DIR = self.modeldirectory
        cfg.merge_from_file(model_zoo.get_config_file("COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml"))
        cfg.DATALOADER.NUM_WORKERS = 2
        cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url("COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml")  # Let training initialize from model zoo
        cfg.SOLVER.IMS_PER_BATCH = 2  # This is the real "batch size" commonly known to deep learning people
        cfg.SOLVER.BASE_LR = 0.00025  # pick a good LR
        cfg.SOLVER.MAX_ITER = 1000    # 1000 iterations seems good enough for this dataset
        cfg.SOLVER.STEPS = []        # do not decay learning rate
        cfg.MODEL.ROI_HEADS.BATCH_SIZE_PER_IMAGE = 256   # Default is 512, using 256 for this dataset.
        cfg.MODEL.ROI_HEADS.NUM_CLASSES = 2 
        cfg.MODEL.WEIGHTS = os.path.join(cfg.OUTPUT_DIR, self.modelweights)  # path to the model we just trained
        cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5   # set a custom testing threshold
        predictor = DefaultPredictor(cfg)

        print("Model Loaded")
        bubInd=np.load(bubind_file, allow_pickle=True)
        frames=np.load(frameind_file, allow_pickle=True)

        def get_image_paths(directory):
            """
            Get a list of file paths for all image files in the specified directory and its subdirectories.

            Args:
            directory (str): The directory to search for image files.

            Returns:
            List[str]: A list of file paths for all image files found, sorted alphabetically.
            """
            image_extensions = ['*.jpg']  # Add more extensions as needed

            image_paths = []
            for extension in image_extensions:
                pattern = os.path.join(directory, '**', extension)
                image_paths.extend(glob.glob(pattern, recursive=True))

            return sorted(image_paths)  # Sort the list of image paths alphabetically




        # Get a list of image file paths sorted alphabetically
        image_paths = get_image_paths(directory_path)

        
        # Set the output video file name
        output_file = f'./vel_bubble{bubble}1.avi'
        frame_width = 832
        frame_height = 600
        '''
        fourcc = cv2.VideoWriter_fourcc(*'XVID')
        out = cv2.VideoWriter(output_file, fourcc, 10, (frame_width, frame_height), isColor=True)
        '''
        skip=5
        angles=[]
        #Save Contours of a bubble in each frame

        image_paths_sub=image_paths[0:]

        #output_video = 'bubble0.avi'
        #frame_size = (832, 600)
        #fourcc = cv2.VideoWriter_fourcc(*'XVID')
        #out = cv2.VideoWriter(output_video, fourcc, 20.0, frame_size, isColor=False)
        #bubble=12
        contours=[]
        centroids=[]
        values=[]
        #for i in tqdm(range(len(frames[bubble]))):
        indexs=[]
        print(len(frames[bubble]))
        for i in tqdm(range(len(frames[bubble]))):
            new_im = cv2.imread(image_paths_sub[frames[bubble][i]])
            outputs = predictor(new_im)
            box = outputs["instances"].pred_boxes
            box = box.tensor
            box = box.cpu().tolist()
            masks = outputs["instances"].pred_masks.cpu()
            scores = outputs["instances"].scores
            scores = scores.cpu().tolist()
            k = 0
            for j in range(len(box)):
                x1, y1, x2, y2 = box[j]
                if y2 > -1000:
                    if k == bubInd[bubble][i]:
                        mask = np.uint8(masks[j]) * 255
                        contours1, _=cv2.findContours(mask, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_NONE)
                        if len(contours1)>1:
                            largest_contour=max(contours1, key=cv2.contourArea)
                            contours1=np.array(largest_contour).reshape((-1,2))
                        else:
                            contours1=np.array(contours1).reshape((-1,2))
                            #out.write(mask)
                        contours.append(contours1)
                        indexs.append(j)
                    k += 1




        avg_mag=[]
        mag=[]
        contour_connections=[]
        for j in range(len(frames[bubble]) - skip-1):

            newimg = cv2.imread(image_paths_sub[frames[bubble][j]])

            frame1 = frames[bubble][j]
            frame2 = frames[bubble][j + skip]
            tree_set1 = cKDTree(contours[j + skip])
            distances, indices = tree_set1.query(contours[j], k=1)
            if j ==1:
                fig, ax = plt.subplots()
                ax.imshow(newimg)
                ax.scatter(contours[j][:, 0], contours[j][:, 1], label='Set 1', marker='o', s=1)
                ax.scatter(contours[j + skip][:, 0], contours[j + skip][:, 1], label='Set 2', marker='x', s=1)
                plt.show()
            
            velocitys=(np.array(distances)/184)/((frame2-frame1)/3000)
            distances=velocitys
            avg_mag.append(np.mean(np.array(distances)))
            mag.append(list(distances))
            # Connect the paired points
            ang=[]
            #print(len(contours[j]), len(indices), max(indices))
            '''
            if j ==0:
                values.append(indices[0])
            else:
                values.append(indices[values[-1]])
            '''  

            for i in range(0, len(contours[j]), 1):
                point1 = contours[j][i]
                point2 = contours[j + skip][indices[i]]
                vector=point1-point2
                angle_rad=np.arctan2(vector[1],vector[0])
                angle_rad=np.degrees(angle_rad)
                if angle_rad <0:
                    angle_rad +=360
                ang.append(angle_rad)
                #ax.arrow(point1[0], point1[1], point2[0] - point1[0], point2[1] - point1[1],
                #         head_width=4, head_length=6, fc='k', ec='k', linewidth=0.5)
            angles.append(ang)
            contour_connections.append(contours[j+skip][indices])

            direction=[]
        for i in range(len(contours)-skip-1):
        #for i in range(1):
            new_im = cv2.imread(image_paths_sub[frames[bubble][i]])
            outputs = predictor(new_im)
            masks = outputs["instances"].pred_masks.cpu()
            mask = np.uint8(masks[indexs[i]]) * 255
            class_val=[]
            for j in range(len(contours[i])):
                x_coord=contour_connections[i][j][0]
                y_coord=contour_connections[i][j][1]
                if x_coord >=832:
                    x_coord =831
                elif x_coord <=0:
                    x_coord=0

                if y_coord <=0:
                    y_coord=0
                elif y_coord >=600:
                    y_coord=599

                if mask[y_coord][x_coord]==255:
                    class_val.append(1)
                else:
                    class_val.append(0)
            direction.append(class_val)


        for i in range(len(direction)):
            for j in range(len(direction[i])):
                if direction[i][j]==1:
                    mag[i][j] = mag[i][j] *-1

        length=len(mag[0])
        for i in range(len(mag)):
            if len(mag[i]) <= length:
                length=len(mag[i])

        num_entries = 200
        data=np.empty((len(mag), num_entries))
        # Calculate indices for evenly spaced entries
        for i in range(len(mag)):
            indices = np.linspace(0, len(mag[i]) - 1, num_entries, dtype=int)
            data[i:]=np.array(mag[i])[indices]
        #data=data[:-20]

        data1=np.empty((len(mag), num_entries))
        #data1=data1[:-20]
        split_val=100
        data1[:,0:-split_val]=data[:,split_val:]
        data1[:,-split_val:]=data[:,0:split_val]

        data=data1

        data_smoothed = gaussian_filter(data, sigma=2)

        # Generate some data
        total_time=len(data)/3000
        # Plot the data with a color bar

        image=plt.imshow(data_smoothed.T,extent=[0,total_time,0,200], aspect='auto', cmap='turbo', alpha=1)
        cbar = plt.colorbar()
        cbar.set_label('Velocity Magnitude (cm/s)')  # Set label for the color bar
        image.set_clim(vmin=-30, vmax=30)  # Set the range of values to display

        plt.xticks(np.arange(0, total_time, 0.01), [str(i) for i in np.arange(0, total_time, 0.01)])

        plt.xlabel('Time (s)')
        plt.ylabel('Location Along Bubble \n Perimeter')
        # Control the range of values displayed on the color bar
        saveloc = self.savefolder + f'velocity_{self.extension}_{bubble}.png'
        plt.savefig(saveloc, bbox_inches='tight')
        plt.show()


